<?php
namespace Uiharu\Lookup;

use finfo;
use stdClass;
use DOMDocument;
use RuntimeException;
use Uiharu\Config;
use Uiharu\FFMPEG;
use Uiharu\MediaTypeExts;
use Uiharu\Url;
use Index\MediaType;

// TODO: Content-Disposition should be honoured for the filename (title).
final class WebLookup implements \Uiharu\ILookup {
    public function match(Url $url): bool {
        return $url->isWeb();
    }

    private static function reqCreate(string $url) {
        $curl = curl_init($url);
        curl_setopt_array($curl, [
            CURLOPT_AUTOREFERER => true,
            CURLOPT_CERTINFO => false,
            CURLOPT_FAILONERROR => false,
            CURLOPT_FOLLOWLOCATION => true,
            CURLOPT_MAXREDIRS => 5,
            CURLOPT_PATH_AS_IS => true,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_TCP_FASTOPEN => true,
            CURLOPT_CONNECTTIMEOUT => 2,
            CURLOPT_PROTOCOLS => CURLPROTO_HTTP | CURLPROTO_HTTPS,
            CURLOPT_REDIR_PROTOCOLS => CURLPROTO_HTTP | CURLPROTO_HTTPS,
            CURLOPT_TIMEOUT => 5,
            CURLOPT_DEFAULT_PROTOCOL => 'https',
            CURLOPT_USERAGENT => 'Mozilla/5.0 (compatible; FlashiiBot/2.5; Uiharu/' . UIH_VERSION . '; +http://fii.moe/uiharu)',
            CURLOPT_HTTPHEADER => [
                'Accept: text/html, application/xhtml+xml, application/xml;q=0.9, */*;q=0.8',
                'Accept-Language: en-GB, en;q=0.9, ja-jp;q=0.6, *;q=0.5',
            ],
        ]);
        return $curl;
    }

    private static function reqHead($curl): array|null {
        curl_setopt_array($curl, [
            CURLOPT_NOBODY => true,
            CURLOPT_HEADER => true,
        ]);

        $headers = curl_exec($curl);
        if($headers === false)
            return null;

        $headers = explode("\r\n", trim($headers));
        $status = 200;
        $lines = [];

        foreach($headers as $header) {
            if(empty($header))
                continue;

            if(strpos($header, ':') === false) {
                $headParts = explode(' ', $header);
                if(isset($headParts[1]) && is_numeric($headParts[1]))
                    $status = (int)$headParts[1];
                $lines = [];
                continue;
            }

            $parts = explode(':', $header, 2);
            $parts[0] = mb_strtolower($parts[0]);
            if(isset($lines[$parts[0]]))
                $lines[$parts[0]] .= ', ' . trim($parts[1] ?? '');
            else
                $lines[$parts[0]] = trim($parts[1] ?? '');
        }

        return compact('status', 'lines');
    }

    private static function reqBody($curl): string|false {
        curl_setopt_array($curl, [
            CURLOPT_NOBODY => false,
            CURLOPT_HEADER => false,
        ]);

        return curl_exec($curl);
    }

    private static function reqError($curl): string {
        return curl_error($curl);
    }

    private static function reqClose($curl): void {
        curl_close($curl);
    }

    public function lookup(Url $url): WebLookupResult {
        $req = self::reqCreate($url);
        $head = self::reqHead($req);

        if($head === null)
            throw new RuntimeException('Web request timed out: ' . self::reqError($req));

        $mediaType = MediaType::parse('application/octet-stream');
        $hasContentType = array_key_exists('content-type', $head['lines']);

        if($hasContentType) {
            try {
                $mediaType = MediaType::parse($head['lines']['content-type'] ?? '');
            } catch(InvalidArgumentException $ex) {}

            if(MediaTypeExts::isMedia($mediaType)) {
                self::reqClose($req);
                return $this->lookupMedia($url, $mediaType);
            }
        }

        $body = self::reqBody($req);
        self::reqClose($req);

        if(!$hasContentType)
            try {
                $finfo = new finfo(FILEINFO_MIME);
                $mediaType = MediaType::parse($finfo->buffer($body));
            } catch(InvalidArgumentException $ex) {}

        if($mediaType->equals('text/html')
            || $mediaType->equals('application/xhtml+xml')
            || $mediaType->equals('application/xml'))
            return $this->lookupSite($url, $req, $mediaType, $body);

        return new WebLookupFallbackResult($url, $mediaType, $url->getHost() . ': ' . basename($url->getPath()));
    }

    private function lookupSite(Url $url, $req, MediaType $mediaType, string $body): WebLookupResult {
        // ok hear me out
        // there's absolutely no good html scraping libraries for PHP
        // DOMDocument Exists but kinda blows at catching weird encoding events like with pixiv
        // and i'm not about to rewrite this whole fucking thing in nodejs
        // also at this point Index should probably provide a wrapper for proc_open lol
        $extract = proc_open(
            sprintf('node %s/extract.mjs', UIH_ROOT),
            [0 => ['pipe', 'r'], 1 => ['pipe', 'w'], 2 => ['pipe', 'w']],
            $pipes
        );
        if(!is_resource($extract))
            throw new RuntimeException('Could not open extract.');

        try {
            fwrite($pipes[0], $body);
            fclose($pipes[0]);

            $stderr = trim(stream_get_contents($pipes[2]));
            if(!empty($stderr))
                throw new RuntimeException('extract: ' . $stderr);

            $stdout = trim(stream_get_contents($pipes[1]));
            if(empty($stdout))
                throw new RuntimeException('extract did not report any errors but exited without any output');
        } finally {
            proc_close($extract);
        }

        $siteInfo = json_decode($stdout);
        if(empty($siteInfo))
            throw new RuntimeException('Failed to parse extract output.');

        return new WebLookupSiteResult($url, $mediaType, $siteInfo);
    }

    private function lookupMedia(Url $url, MediaType $mediaType): WebLookupResult {
        $mediaInfo = FFMPEG::cleanProbe($url);
        return new WebLookupMediaResult($url, $mediaType, $mediaInfo);
    }
}
